import pytest
from django.contrib import admin

from data_browser import migration_helpers
from data_browser.models import View
from data_browser.orm_admin import get_models
from data_browser.orm_results import get_results
from data_browser.query import BoundQuery, Query

from .conftest import ARRAY_FIELD_SUPPORT

if ARRAY_FIELD_SUPPORT:  # pragma: postgres
    from .array.models import ArrayModel
else:  # pragma: not postgres
    pytestmark = pytest.mark.skip("Needs ArrayField support")


class ArrayAdmin(admin.ModelAdmin):
    fields = [
        "char_array_field",
        "int_array_field",
        "char_choice_array_field",
        "int_choice_array_field",
    ]


@pytest.fixture
def with_arrays(db):  # pragma: postgres
    admin.site.register(ArrayModel, ArrayAdmin)
    yield
    admin.site.unregister(ArrayModel)


@pytest.fixture
def get_results_flat(with_arrays, req):  # pragma: postgres
    def helper(fields, query=None):
        query = query or {}

        orm_models = get_models(req)
        query = Query.from_request("array.ArrayModel", fields, query)
        bound_query = BoundQuery.bind(query, orm_models)
        data = get_results(req, bound_query, orm_models, False)

        for f in bound_query.filters:
            if f.err_message:
                print(
                    "filter error:", f.path_str, f.lookup, f.value, "->", f.err_message
                )

        return data["rows"]

    return helper


def test_hello_world(get_results_flat):  # pragma: postgres
    ArrayModel.objects.create(
        int_array_field=[1, 2],
        char_array_field=["a", "b"],
        int_choice_array_field=[1, 2],
        char_choice_array_field=["a", "b"],
    )
    assert get_results_flat(
        "int_array_field,char_array_field,int_choice_array_field,char_choice_array_field"
    ) == [
        {
            "int_array_field": "[1.0, 2.0]",
            "char_array_field": '["a", "b"]',
            "int_choice_array_field": '["A", "B"]',
            "char_choice_array_field": '["A", "B"]',
        }
    ]


def test_int_array_contains(get_results_flat):  # pragma: postgres
    ArrayModel.objects.create(int_array_field=[1, 2])
    ArrayModel.objects.create(int_array_field=[2, 3])
    ArrayModel.objects.create(int_array_field=[1, 3])
    assert get_results_flat("int_array_field", {"int_array_field__contains": [2]}) == [
        {"int_array_field": "[1.0, 2.0]"},
        {"int_array_field": "[2.0, 3.0]"},
    ]


def test_int_choice_array_contains(get_results_flat):  # pragma: postgres
    ArrayModel.objects.create(int_choice_array_field=[1, 2])
    ArrayModel.objects.create(int_choice_array_field=[2, 3])
    ArrayModel.objects.create(int_choice_array_field=[1, 3])
    assert get_results_flat(
        "int_choice_array_field", {"int_choice_array_field__contains": ["B"]}
    ) == [
        {"int_choice_array_field": '["A", "B"]'},
        {"int_choice_array_field": '["B", "C"]'},
    ]


def test_char_array_contains(get_results_flat):  # pragma: postgres
    ArrayModel.objects.create(char_array_field=["a", "b"])
    ArrayModel.objects.create(char_array_field=["b", "c"])
    ArrayModel.objects.create(char_array_field=["a", "c"])
    assert get_results_flat(
        "char_array_field", {"char_array_field__contains": ["b"]}
    ) == [{"char_array_field": '["a", "b"]'}, {"char_array_field": '["b", "c"]'}]


def test_char_choice_array_contains(get_results_flat):  # pragma: postgres
    ArrayModel.objects.create(char_choice_array_field=["a", "b"])
    ArrayModel.objects.create(char_choice_array_field=["b", "c"])
    ArrayModel.objects.create(char_choice_array_field=["a", "c"])
    assert get_results_flat(
        "char_choice_array_field", {"char_choice_array_field__contains": ["B"]}
    ) == [
        {"char_choice_array_field": '["A", "B"]'},
        {"char_choice_array_field": '["B", "C"]'},
    ]


def test_filter_length(get_results_flat):  # pragma: postgres
    ArrayModel.objects.create(int_array_field=[1])
    ArrayModel.objects.create(int_array_field=[1, 2])
    ArrayModel.objects.create(int_array_field=[1, 2, 3])
    assert get_results_flat("int_array_field", {"int_array_field__length": [2]}) == [
        {"int_array_field": "[1.0, 2.0]"}
    ]


def test_choice_array_filter_length(get_results_flat):  # pragma: postgres
    ArrayModel.objects.create(int_choice_array_field=[1])
    ArrayModel.objects.create(int_choice_array_field=[1, 2])
    ArrayModel.objects.create(int_choice_array_field=[1, 2, 3])
    assert get_results_flat(
        "int_choice_array_field", {"int_choice_array_field__length": [2]}
    ) == [{"int_choice_array_field": '["A", "B"]'}]


def test_char_choice_array_equals(get_results_flat):  # pragma: postgres
    ArrayModel.objects.create(char_choice_array_field=["a", "b"])
    ArrayModel.objects.create(char_choice_array_field=["b", "c"])
    ArrayModel.objects.create(char_choice_array_field=["a", "c"])
    assert get_results_flat(
        "char_choice_array_field", {"char_choice_array_field__equals": ['["A", "B"]']}
    ) == [{"char_choice_array_field": '["A", "B"]'}]


@pytest.mark.parametrize(
    "before,after",
    [
        # choice_array contains / not_contains
        ("char_choice_array_field__contains=a", "char_choice_array_field__contains=A"),
        (
            "char_choice_array_field__contains=d",
            "char_choice_array_field__raw__contains=d",
        ),
        ("int_choice_array_field__contains=1", "int_choice_array_field__contains=A"),
        (
            "int_choice_array_field__contains=4",
            "int_choice_array_field__raw__contains=4.0",
        ),
        # choice_array other -> noop
        ("char_choice_array_field__length=1", "char_choice_array_field__length=1"),
        ("int_choice_array_field__length=1", "int_choice_array_field__length=1"),
        ("char_choice_array_field__wtf=1", "char_choice_array_field__wtf=1"),
        ("int_choice_array_field__wtf=1", "int_choice_array_field__wtf=1"),
        # regular array -> noop
        ("char_array_field__contains=a", "char_array_field__contains=a"),
        ("int_array_field__contains=1", "int_array_field__contains=1"),
    ],
)
def test_0009(req, with_arrays, before, after):  # pragma: postgres
    orm_models = get_models(req)
    valid = int("wtf" not in before)

    view = View.objects.create(model_name="array.ArrayModel", query=before)
    migration_helpers.forwards_0009(View)
    view.refresh_from_db()
    assert view.query == after
    assert len(BoundQuery.bind(view.get_query(), orm_models).valid_filters) == valid



import django
import pytest
from django.contrib import admin

from data_browser.helpers import AdminMixin
from data_browser.orm_admin import get_models
from data_browser.orm_results import get_results
from data_browser.query import BoundQuery, Query

from .conftest import JSON_FIELD_SUPPORT, SQLITE

""" Howto enable SQLite JSON support https://code.djangoproject.com/wiki/JSON1Extension """


if JSON_FIELD_SUPPORT:
    from .json.models import JsonModel
else:  # pragma: no cover
    pytestmark = pytest.mark.skip("Needs JSONField support")


class JsonAdmin(AdminMixin, admin.ModelAdmin):
    fields = ["json_field"]
    ddb_json_fields = {
        "json_field": {"hello": "string", "position": "number", "bool": "boolean"}
    }


@pytest.fixture
def with_json(db):
    admin.site.register(JsonModel, JsonAdmin)
    yield
    admin.site.unregister(JsonModel)


@pytest.fixture
def get_results_flat(with_json, req):
    def helper(fields, query=None):
        query = query or {}

        orm_models = get_models(req)
        query = Query.from_request("json.JsonModel", fields, query)
        bound_query = BoundQuery.bind(query, orm_models)
        data = get_results(req, bound_query, orm_models, False)

        for f in bound_query.filters:
            if f.err_message:
                print(
                    "filter error:", f.path_str, f.lookup, f.value, "->", f.err_message
                )

        return data["rows"]

    return helper


def test_hello_world(get_results_flat):
    JsonModel.objects.create(json_field={"hello": "world"})
    assert get_results_flat("json_field") == [{"json_field": '{"hello": "world"}'}]


def test_get_string_sub_field(get_results_flat):
    JsonModel.objects.create(json_field={"hello": "world"})
    assert get_results_flat("json_field__hello") == [{"json_field__hello": "world"}]


@pytest.mark.skipif(
    SQLITE and django.VERSION[:3] == (3, 1, 3),
    reason="https://code.djangoproject.com/ticket/32203",
)
def test_get_number_sub_field(get_results_flat):  # pragma: not sqlite
    JsonModel.objects.create(json_field={"position": 1})
    assert get_results_flat("json_field__position") == [{"json_field__position": 1}]


@pytest.mark.skipif(
    SQLITE and django.VERSION[:3] == (3, 1, 3),
    reason="https://code.djangoproject.com/ticket/32203",
)
def test_get_boolean_sub_field(get_results_flat):  # pragma: not sqlite
    JsonModel.objects.create(json_field={"bool": True})
    assert get_results_flat("json_field__bool") == [{"json_field__bool": True}]


@pytest.mark.skipif(
    SQLITE and django.VERSION[:3] == (3, 1, 3),
    reason="https://code.djangoproject.com/ticket/32203",
)
def test_sub_field_is_null(get_results_flat):  # pragma: not sqlite
    JsonModel.objects.create(json_field={"position": 1, "hello": "world"})
    JsonModel.objects.create(json_field={"position": 2, "hello": None})
    JsonModel.objects.create(json_field={"position": 3, "goodbye": "world"})
    assert get_results_flat("json_field__hello__is_null,json_field__position+1") == [
        {"json_field__hello__is_null": "NotNull", "json_field__position": 1},
        {"json_field__hello__is_null": "IsNull", "json_field__position": 2},
        {"json_field__hello__is_null": "IsNull", "json_field__position": 3},
    ]

    # __is_null=
    assert get_results_flat(
        "json_field__position+1", {"json_field__hello__is_null": ["NotNull"]}
    ) == [{"json_field__position": 1}]
    assert get_results_flat(
        "json_field__position+1", {"json_field__hello__is_null": ["IsNull"]}
    ) == [{"json_field__position": 2}, {"json_field__position": 3}]

    # __is_null__equals=
    assert get_results_flat(
        "json_field__position+1", {"json_field__hello__is_null__equals": ["NotNull"]}
    ) == [{"json_field__position": 1}]
    assert get_results_flat(
        "json_field__position+1", {"json_field__hello__is_null__equals": ["IsNull"]}
    ) == [{"json_field__position": 2}, {"json_field__position": 3}]


def test_filter_sub_field(get_results_flat):
    JsonModel.objects.create(json_field={"hello": "world"})
    JsonModel.objects.create(json_field={"hello": "universe"})
    assert get_results_flat("json_field", {"json_field__hello__equals": ["world"]}) == [
        {"json_field": '{"hello": "world"}'}
    ]


def test_filter_field_value(get_results_flat):
    JsonModel.objects.create(json_field={"goodbye": "universe"})
    assert get_results_flat("json_field") == [{"json_field": '{"goodbye": "universe"}'}]
    assert (
        get_results_flat(
            "json_field", {"json_field__field_equals": ['goodbye|"world"']}
        )
        == []
    )


def test_filter_field_value_no_seperator(get_results_flat):
    JsonModel.objects.create(json_field={"goodbye": "universe"})
    assert get_results_flat(
        "json_field", {"json_field__field_equals": ['goodbye"world"']}
    ) == [{"json_field": '{"goodbye": "universe"}'}]


def test_filter_field_value_no_field(get_results_flat):
    JsonModel.objects.create(json_field={"goodbye": "universe"})
    assert get_results_flat(
        "json_field", {"json_field__field_equals": ['|"world"']}
    ) == [{"json_field": '{"goodbye": "universe"}'}]


def test_filter_field_value_list(get_results_flat):
    JsonModel.objects.create(json_field={"goodbye": ["world"]})
    JsonModel.objects.create(json_field={"goodbye": ["universe"]})
    assert get_results_flat(
        "json_field", {"json_field__field_equals": ['goodbye|["world"]']}
    ) == [{"json_field": '{"goodbye": ["world"]}'}]


def test_filter_field_value_bad_json(get_results_flat):
    JsonModel.objects.create(json_field={"goodbye": "universe"})
    assert get_results_flat(
        "json_field", {"json_field__field_equals": ["goodbye|world"]}
    ) == [{"json_field": '{"goodbye": "universe"}'}]


def test_filter_has_key(get_results_flat):
    JsonModel.objects.create(json_field={"hello": "world"})
    JsonModel.objects.create(json_field={"goodbye": "world"})
    assert get_results_flat("json_field", {"json_field__has_key": ["hello"]}) == [
        {"json_field": '{"hello": "world"}'}
    ]


def test_filter_equals(get_results_flat):
    JsonModel.objects.create(json_field={"hello": "world"})
    JsonModel.objects.create(json_field={"goodbye": "world"})
    assert get_results_flat(
        "json_field", {"json_field__equals": ['{"hello": "world"}']}
    ) == [{"json_field": '{"hello": "world"}'}]


def test_filter_equals_bad_json(get_results_flat):
    JsonModel.objects.create(json_field={"hello": "world"})
    JsonModel.objects.create(json_field={"goodbye": "world"})
    assert get_results_flat(
        "json_field-1", {"json_field__equals": ['{"hello": "world"']}
    ) == [{"json_field": '{"hello": "world"}'}, {"json_field": '{"goodbye": "world"}'}]



import json
from collections import defaultdict
from dataclasses import dataclass

from django.contrib.admin import site
from django.contrib.admin.options import BaseModelAdmin
from django.contrib.admin.utils import flatten_fieldsets
from django.contrib.auth.admin import UserAdmin
from django.db import models
from django.db.models.fields.reverse_related import ForeignObjectRel, OneToOneRel
from django.forms.models import _get_foreign_key
from django.urls import reverse
from django.utils.html import format_html
from django.utils.text import slugify

from .common import debug_log, settings
from .helpers import AdminMixin, _AnnotationDescriptor, _get_option, attributes
from .orm_aggregates import get_aggregates_for_type
from .orm_fields import (
    OrmAnnotatedField,
    OrmCalculatedField,
    OrmConcreteField,
    OrmFileField,
    OrmFkField,
    OrmRawField,
)
from .orm_functions import get_functions_for_type
from .orm_types import get_field_type
from .types import TYPES, BooleanType, JSONType, NumberType, StringType

OPEN_IN_ADMIN = "admin"


@dataclass
class OrmModel:
    fields: dict
    admin: BaseModelAdmin = None

    @property
    def root(self):
        return bool(self.admin)

    @property
    def default_filters(self):
        default_filters = _get_option(self.admin, "default_filters")
        assert isinstance(default_filters, list)
        return [
            (f, l, v if isinstance(v, str) else json.dumps(v))
            for (f, l, v) in default_filters
        ]


def get_model_name(model, sep="."):
    return f"{model._meta.app_label}{sep}{model.__name__}"


def get_fields_for_type(type_):
    aggregates = get_aggregates_for_type(type_)
    functions = get_functions_for_type(type_)
    others = {}
    if type_.raw_type:
        others["raw"] = OrmRawField(
            type_.name, "raw", "raw", type_.raw_type, type_.raw_type.name, None
        )

    return {**aggregates, **functions, **others}


@attributes(short_description=OPEN_IN_ADMIN)
def open_in_admin(obj):
    if obj is None:  # pragma: no cover
        return None

    model_name = get_model_name(obj.__class__, "_")
    url_name = f"admin:{model_name}_change".lower()
    url = reverse(url_name, args=[obj.pk])
    return format_html('<a href="{}">{}</a>', url, obj)


def admin_get_queryset(admin, request, fields=()):
    request.data_browser = {"calculated_fields": set(fields), "fields": set(fields)}
    return admin.get_queryset(request)


def _get_all_admin_fields(request):
    request.data_browser = {"calculated_fields": set(), "fields": set()}

    def from_fieldsets(admin, all_):
        auth_user_compat = settings.DATA_BROWSER_AUTH_USER_COMPAT
        if auth_user_compat and isinstance(admin, UserAdmin):
            obj = admin.model()  # get the change fieldsets, not the add ones
        else:
            obj = None

        fields = admin.get_fieldsets(request, obj)
        for f in flatten_fieldsets(fields):
            # skip calculated fields on inlines
            if all_ or hasattr(admin.model, f):
                yield f

    def visible(model_admin, request):
        has_attrs = all(
            hasattr(model_admin, a) for a in ["get_fieldsets", "model", "get_queryset"]
        )
        if not has_attrs or not issubclass(model_admin.model, models.Model):
            debug_log(
                f"{type(model_admin)} instance does not look like a ModelAdmin or InlineModelAdmin"
            )
            return False

        if _get_option(model_admin, "ignore", request):
            return False

        if model_admin.has_change_permission(request):
            return True

        if hasattr(model_admin, "has_view_permission"):
            return model_admin.has_view_permission(request)
        else:
            return False  # pragma: no cover  Django < 2.1

    all_admin_fields = defaultdict(set)
    hidden_fields = defaultdict(set)

    def get_common(admin, model):
        all_admin_fields[model].update(from_fieldsets(admin, False))
        all_admin_fields[model].update(_get_option(admin, "extra_fields", request))
        hidden_fields[model].update(_get_option(admin, "hide_fields", request))

    model_admins = {}
    for model, model_admin in site._registry.items():
        if visible(model_admin, request):
            model_admins[model] = model_admin
            all_admin_fields[model].update(model_admin.get_list_display(request))
            all_admin_fields[model].add(open_in_admin)
            get_common(model_admin, model)

            # check the inlines, these are already filtered for access
            for inline in model_admin.get_inline_instances(request):
                if visible(inline, request):
                    try:
                        fk_field = _get_foreign_key(model, inline.model, inline.fk_name)
                    except Exception as e:
                        debug_log(e)  # ignore things like GenericInlineModelAdmin
                    else:
                        if inline.model not in model_admins:  # pragma: no branch
                            model_admins[inline.model] = inline
                        all_admin_fields[inline.model].add(fk_field.name)
                        get_common(inline, inline.model)

    for model, model_admin in model_admins.items():
        if isinstance(model_admin, AdminMixin):
            all_admin_fields[model].update(model_admin._ddb_annotations())
        all_admin_fields[model].update(from_fieldsets(model_admin, True))

    # we always have id and never pk
    for fields in all_admin_fields.values():
        fields.add("id")
        fields.discard("pk")
        fields.discard("__str__")

    # throw away the hidden ones
    for model, fields in hidden_fields.items():
        for f in fields:
            all_admin_fields[model].discard(f)

    return model_admins, all_admin_fields


def _upper(name):
    return name[0].upper() + name[1:]


def _get_calculated_field(request, field_name, model_name, model, admin, model_fields):
    if isinstance(field_name, str):
        field_func = getattr(admin, field_name, None)
    else:
        # an actual function in list_display
        field_func = field_name
        field_name = slugify(
            getattr(field_func, "short_description", str(field_func.__name__))
        )

    if getattr(field_func, "ddb_hide", False):
        return None

    pretty_name = _upper(
        getattr(field_func, "short_description", field_name.replace("_", " "))
    )

    if isinstance(field_func, _AnnotationDescriptor):
        qs = admin_get_queryset(admin, request, [field_name])

        annotation = qs.query.annotations.get(field_name)
        if not annotation:  # pragma: no cover
            raise Exception(
                f"Can't find annotation '{field_name}' for {admin}.{field_name}"
            )

        field = getattr(annotation, "output_field", None)
        if not field:  # pragma: no cover
            raise Exception(
                f"Annotation '{field_name}' for {admin}.{field_name} doesn't specify 'output_field'"
            )

        type_, choices = get_field_type(field)
        return OrmAnnotatedField(
            model_name=model_name,
            name=field_name,
            pretty_name=pretty_name,
            type_=type_,
            django_field=field,
            admin=admin,
            choices=choices,
        )
    else:
        if field_func is None:

            def field_func(obj):
                value = getattr(obj, field_name)
                return value() if callable(value) else value

        return OrmCalculatedField(
            model_name=model_name,
            name=field_name,
            pretty_name=pretty_name,
            func=field_func,
        )


def _make_json_sub_module(model_name, field_types):
    TYPE_MAP = {"string": StringType, "number": NumberType, "boolean": BooleanType}

    fields = dict(get_fields_for_type(JSONType))
    for field_name, type_name in field_types.items():
        type_ = TYPE_MAP[type_name]
        fields[field_name] = OrmConcreteField(
            model_name=model_name,
            name=field_name,
            pretty_name=field_name,
            type_=type_,
            rel_name=type_.name,
            choices=None,
        )

    return OrmModel(fields)


def _get_fields_for_model(request, model, admin, admin_fields):
    fields = {}
    orm_models = {}

    model_name = get_model_name(model)
    model_fields = {f.name: f for f in model._meta.get_fields()}

    for field_name in admin_fields[model]:
        field = model_fields.get(field_name)
        pretty_name = (
            _upper(getattr(field, "verbose_name", field_name)) if field else None
        )
        # FK's and OneToOne's
        if isinstance(field, (models.ForeignKey, OneToOneRel)):
            if field.related_model in admin_fields:
                fields[field_name] = OrmFkField(
                    model_name=model_name,
                    name=field_name,
                    pretty_name=pretty_name,
                    rel_name=get_model_name(field.related_model),
                )
        # ManyToMany
        elif isinstance(field, (ForeignObjectRel, models.ManyToManyField)):
            pass
        # Files and Images
        elif isinstance(field, models.FileField):
            fields[field_name] = OrmFileField(
                model_name=model_name,
                name=field_name,
                pretty_name=pretty_name,
                django_field=field,
            )
        # Calculated and annoted fields
        elif isinstance(field, type(None)):
            orm_field = _get_calculated_field(
                request, field_name, model_name, model, admin, model_fields
            )
            if orm_field:
                fields[orm_field.name] = orm_field
        # Normal fields
        else:
            field_type, choices = get_field_type(field)

            rel_name = field_type.name
            if field_type is JSONType:
                json_fields = _get_option(admin, "json_fields", request).get(field_name)
                if json_fields:  # pragma: no branch
                    rel_name = f"{model_name}__{field_name}"
                    orm_models[rel_name] = _make_json_sub_module(
                        model_name, json_fields
                    )

            fields[field_name] = OrmConcreteField(
                model_name=model_name,
                name=field_name,
                pretty_name=pretty_name,
                type_=field_type,
                rel_name=rel_name,
                choices=choices,
            )
    orm_models[model_name] = OrmModel(fields=fields, admin=admin)
    return orm_models


def get_models(request):
    model_admins, admin_fields = _get_all_admin_fields(request)
    models = {}
    for model in admin_fields:
        models.update(
            _get_fields_for_model(request, model, model_admins[model], admin_fields)
        )
    types = {
        type_.name: OrmModel(get_fields_for_type(type_)) for type_ in TYPES.values()
    }

    return {**models, **types}



from django.contrib.postgres.fields import ArrayField  # pragma: postgres
from django.db import models  # pragma: postgres


class ArrayModel(models.Model):  # pragma: postgres
    char_array_field = ArrayField(null=True, base_field=models.CharField(max_length=32))
    int_array_field = ArrayField(null=True, base_field=models.IntegerField())
    char_choice_array_field = ArrayField(
        null=True,
        base_field=models.CharField(
            max_length=32, choices=[("a", "A"), ("b", "B"), ("c", "C")]
        ),
    )
    int_choice_array_field = ArrayField(
        null=True,
        base_field=models.IntegerField(choices=[(1, "A"), (2, "B"), (3, "C")]),
    )



import json
from pathlib import Path


class ANY:
    def __init__(self, type):
        self.type = type

    def __eq__(self, other):
        return isinstance(other, self.type)


class KEYS:
    def __init__(self, *keys):
        self.keys = set(keys)

    def __eq__(self, other):
        return isinstance(other, dict) and other.keys() == self.keys


def update_fe_fixture(filename, data):  # pragma: no cover
    filename = Path(filename)

    if filename.exists():
        with filename.open("r") as f:
            current = json.load(f)
    else:
        current = None

    if data != current:
        with filename.open("w") as f:
            json.dump(data, f, indent=4, sort_keys=True)
            f.write("\n")



from django.contrib import admin
from django.contrib.admin.utils import flatten_fieldsets
from django.utils.html import format_html

from . import models
from .common import can_make_public
from .helpers import AdminMixin


@admin.register(models.View)
class ViewAdmin(AdminMixin, admin.ModelAdmin):
    fieldsets = [
        (None, {"fields": ["name", "owner", "open_view", "description"]}),
        (
            "Public",
            {
                "fields": [
                    "public",
                    "public_slug",
                    "public_link",
                    "google_sheets_formula",
                ]
            },
        ),
        ("Query", {"fields": ["model_name", "fields", "query", "limit"]}),
        ("Internal", {"fields": ["id", "created_time"]}),
    ]
    list_display = ["__str__", "owner", "public"]

    def get_readonly_fields(self, request, obj):
        return flatten_fieldsets(self.get_fieldsets(request, obj))

    def get_fieldsets(self, request, obj=None):
        res = super().get_fieldsets(request, obj)
        if not can_make_public(request.user):
            res = [fs for fs in res if fs[0] != "Public"]
        return res

    def change_view(self, request, *args, **kwargs):
        models.global_data.request = request
        return super().change_view(request, *args, **kwargs)

    @staticmethod
    def open_view(obj):
        if not obj.model_name:
            return "N/A"
        url = obj.get_query().get_url("html")
        return format_html(f'<a href="{url}">view</a>')

    def get_changeform_initial_data(self, request):
        get_results = super().get_changeform_initial_data(request)
        get_results["owner"] = request.user.pk
        return get_results



import threading

from django.db import models
from django.http import QueryDict
from django.urls import reverse
from django.utils import crypto, timezone

from .common import MAKE_PUBLIC_CODENAME, settings

global_data = threading.local()


def get_id():
    return crypto.get_random_string(length=12)


class View(models.Model):
    class Meta:
        permissions = [
            (MAKE_PUBLIC_CODENAME, "Can make a saved view publicly available")
        ]

    id = models.CharField(primary_key=True, max_length=12, default=get_id)
    created_time = models.DateTimeField(default=timezone.now)

    name = models.CharField(max_length=64, blank=False)
    description = models.TextField(blank=True)
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True, blank=True, on_delete=models.SET_NULL
    )

    public = models.BooleanField(default=False)
    public_slug = models.CharField(max_length=12, default=get_id, blank=False)

    model_name = models.CharField(max_length=32, blank=False)
    fields = models.TextField(blank=True)
    query = models.TextField(blank=True)
    limit = models.IntegerField(blank=False, null=False, default=1000)

    def get_query(self):
        from .query import Query

        return Query.from_request(self.model_name, self.fields, QueryDict(self.query))

    def public_link(self):
        if self.public:
            if settings.DATA_BROWSER_ALLOW_PUBLIC:
                url = reverse(
                    "data_browser:view", kwargs={"pk": self.public_slug, "media": "csv"}
                )
                return global_data.request.build_absolute_uri(url)
            else:
                return "Public Views are disabled in Django settings."
        else:
            return "N/A"

    def google_sheets_formula(self):
        if self.public:
            if settings.DATA_BROWSER_ALLOW_PUBLIC:
                url = reverse(
                    "data_browser:view", kwargs={"pk": self.public_slug, "media": "csv"}
                )
                url = global_data.request.build_absolute_uri(url)
                return f'=importdata("{url}")'
            else:
                return "Public Views are disabled in Django settings."
        else:
            return "N/A"

    def __str__(self):
        return f"{self.model_name} view: {self.name}"



import os
from pathlib import Path

import setuptools

from data_browser import version

root = Path("data_browser")
data_files = []
for directory in ("fe_build", "templates", "web_root"):
    for (path, _, filenames) in os.walk(root / directory):
        for filename in filenames:
            data_files.append(os.path.join("..", path, filename))


setuptools.setup(
    name="django-data-browser",
    version=version,
    author="Gordon Wrigley",
    author_email="gordon.wrigley@gmail.com",
    description=(
        "A Django app for interactive user friendly browsing "
        "of a Django project DBs."
    ),
    long_description=Path("README.rst").read_text(),
    long_description_content_type="text/x-rst",
    url="https://github.com/tolomea/django-data-browser",
    packages=setuptools.find_packages(exclude=["tests*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    package_data={"": data_files},
    install_requires=[
        "Django>=2.0",
        "python-dateutil",
        'dataclasses; python_version<"3.7"',
        "sqlparse",
    ],
)



import json
import logging
from urllib.parse import urlencode

from django.db.models import BooleanField
from django.urls import reverse


def attributes(**kwargs):
    def inner(func):
        for k, v in kwargs.items():
            setattr(func, k, v)
        return func

    return inner


class _Everything:
    def __contains__(self, item):
        return True


class _AdminOptions:
    ddb_ignore = False
    ddb_extra_fields = []
    ddb_hide_fields = []
    ddb_json_fields = {}
    ddb_default_filters = []

    def get_ddb_ignore(self, request):
        return self.ddb_ignore

    def get_ddb_extra_fields(self, request):
        return self.ddb_extra_fields

    def get_ddb_hide_fields(self, request):
        return self.ddb_hide_fields

    def get_ddb_json_fields(self, request):
        return self.ddb_json_fields

    def get_ddb_default_filters(self):
        return self.ddb_default_filters


def _get_option(admin, name, *args):
    field = f"ddb_{name}"
    func = f"get_ddb_{name}"

    if hasattr(admin, func):
        return getattr(admin, func)(*args)
    else:
        return getattr(admin, field, getattr(_AdminOptions, field))


class _AdminAnnotations:
    def _get_fields_for_request(self, request):
        if hasattr(request, "data_browser"):
            return request.data_browser["fields"]
        elif (
            request.resolver_match
            and request.resolver_match.func.__name__ == "changelist_view"
        ):
            return set(self.get_list_display(request))
        else:
            return _Everything()

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        fields = self._get_fields_for_request(request)

        for name, descriptor in self._ddb_annotations().items():
            if name not in fields:
                continue

            qs = descriptor.get_queryset(self, request, qs)

            annotation = qs.query.annotations.get(descriptor.name)
            if not annotation:  # pragma: no cover
                raise Exception(
                    f"Can't find annotation '{descriptor.name}' for {self}.{descriptor.name}"
                )

            field_type = getattr(annotation, "output_field", None)
            if not field_type:  # pragma: no cover
                raise Exception(
                    f"Annotation '{descriptor.name}' for {self}.{descriptor.name} doesn't specify 'output_field'"
                )

            descriptor.boolean = isinstance(field_type, BooleanField)
        return qs

    def get_readonly_fields(self, request, obj=None):
        res = super().get_readonly_fields(request, obj)
        return list(res) + list(self._ddb_annotations())

    @classmethod
    def _ddb_annotations(cls):
        if not hasattr(cls, "_ddb_annotations_real"):
            cls._ddb_annotations_real = {}
        return cls._ddb_annotations_real


class AdminMixin(_AdminOptions, _AdminAnnotations):
    def changelist_view(self, request, extra_context=None):
        """ Inject ddb_url """
        extra_context = extra_context or {}

        if not self.get_ddb_ignore(request):
            url = reverse(
                "data_browser:query_html",
                args=[f"{self.model._meta.app_label}.{self.model.__name__}", ""],
            )
            args = self.get_ddb_default_filters()
            params = urlencode(
                [
                    (
                        f"{field}__{lookup}",
                        value if isinstance(value, str) else json.dumps(value),
                    )
                    for field, lookup, value in args
                ]
            )
            extra_context["ddb_url"] = f"{url}?{params}"

        return super().changelist_view(request, extra_context)


class _AnnotationDescriptor:
    def __init__(self, get_queryset):
        self.get_queryset = get_queryset

    def __set_name__(self, owner, name):
        self.name = name
        self.__name__ = name
        self.admin_order_field = name
        if not issubclass(owner, AdminMixin):  # pragma: no cover
            raise Exception(
                "Django Data Browser 'annotation' decorator used without 'AdminMixin'"
            )
        owner._ddb_annotations()[name] = self

    def __get__(self, instance, owner=None):
        return self

    def __call__(self, obj):
        return getattr(obj, self.name)

    def __getattr__(self, name):
        return getattr(self.get_queryset, name)


def ddb_hide(func):  # pragma: no cover
    logging.getLogger(__name__).warning(
        "ddb_hide is deprecated in favor of @attributes(ddb_hide=True)"
    )
    func.ddb_hide = True
    return func


annotation = _AnnotationDescriptor



from dataclasses import dataclass

from django import forms
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils import timezone
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _


@deconstructible
@dataclass
class Word:
    name: str

    def __str__(self):
        return self.name


class FakeField(models.Field):  # pragma: no cover
    description = "Text"

    def get_internal_type(self):
        return "TextField"

    def to_python(self, value):
        return value

    def get_prep_value(self, value):
        value = super().get_prep_value(value)
        return self.to_python(value)

    def from_db_value(self, value, expression, connection):
        if value:
            return value
        return {}

    def formfield(self, **kwargs):
        return super().formfield(
            **{
                "max_length": self.max_length,
                **({} if self.choices else {"widget": forms.Textarea}),
                **kwargs,
            }
        )


# models for perm testing


class InAdmin(models.Model):
    name = models.TextField()


class NotInAdmin(models.Model):
    name = models.TextField()


class InlineAdmin(models.Model):
    name = models.TextField()
    in_admin = models.ForeignKey(InAdmin, on_delete=models.CASCADE)


class Ignored(models.Model):
    name = models.TextField()
    in_admin = models.ForeignKey(InAdmin, on_delete=models.CASCADE)


class GenericInlineAdmin(models.Model):
    name = models.TextField()
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    in_admin = GenericForeignKey("content_type", "object_id")


class Normal(models.Model):
    name = models.TextField()
    in_admin = models.ForeignKey(InAdmin, on_delete=models.CASCADE)
    not_in_admin = models.ForeignKey(NotInAdmin, on_delete=models.CASCADE)
    inline_admin = models.ForeignKey(InlineAdmin, on_delete=models.CASCADE)


# general models


class Tag(models.Model):
    name = models.TextField()


class Address(models.Model):
    city = models.TextField()
    street = models.TextField()

    def fred(self):
        assert self.street != "bad", self.street
        return "fred"

    @property
    def tom(self):
        assert self.street != "bad", self.street
        return "tom"


class Producer(models.Model):
    address = models.OneToOneField(Address, on_delete=models.CASCADE, null=True)
    name = models.TextField()


class Product(models.Model):
    name = models.TextField()
    producer = models.ForeignKey(Producer, on_delete=models.CASCADE)
    size = models.IntegerField(default=0)
    size_unit = models.TextField()
    default_sku = models.ForeignKey(
        "SKU", null=True, on_delete=models.CASCADE, related_name="products"
    )
    tags = models.ManyToManyField(Tag)
    onsale = models.BooleanField(null=True)
    image = models.FileField()
    fake = FakeField()
    created_time = models.DateTimeField(default=timezone.now)
    only_in_list_view = models.TextField()
    hidden_inline = models.TextField()
    hidden_model = models.TextField()
    extra_inline = models.TextField()
    extra_model = models.TextField()
    boat = models.FloatField(null=True)
    duration = models.DurationField(null=True)
    date = models.DateField(null=True)

    not_in_admin = models.TextField()
    fk_not_in_admin = models.ForeignKey(InAdmin, null=True, on_delete=models.CASCADE)
    model_not_in_admin = models.ForeignKey(
        NotInAdmin, null=True, on_delete=models.CASCADE
    )
    string_choice = models.CharField(
        max_length=8, choices=[("a", _("A")), ("b", Word("B"))]
    )
    number_choice = models.IntegerField(
        choices=[(1, _("A")), (2, Word("B"))], default=1
    )

    def is_onsale(self):
        return False


class SKU(models.Model):
    name = models.TextField()
    product = models.ForeignKey(Product, on_delete=models.CASCADE)



from django.db import models

try:
    from django.db.models import JSONField
except ImportError:  # pragma: django < 3.1
    from django.contrib.postgres.fields import JSONField


class JsonModel(models.Model):
    json_field = JSONField()



from django.contrib import admin
from django.contrib.contenttypes.admin import GenericInlineModelAdmin
from django.db.models import F

from data_browser.helpers import AdminMixin, annotation, attributes

from . import models

# admin for perm testing


class InlineAdminInline(admin.TabularInline):
    model = models.InlineAdmin
    fields = ["name"]


class GenericInlineAdminInline(GenericInlineModelAdmin):
    model = models.GenericInlineAdmin


class IgnoredAdminInline(admin.TabularInline):
    model = models.Ignored
    fields = ["name"]
    ddb_ignore = True


class NotAnAdminInline(admin.TabularInline):
    model = models.Ignored
    fields = ["name"]

    @property
    def get_queryset(self):
        raise AttributeError


@admin.register(models.Ignored)
class IgnoredAdmin(AdminMixin, admin.ModelAdmin):
    fields = ["name", "in_admin"]
    ddb_ignore = True


@admin.register(models.InAdmin)
class InAdmin(admin.ModelAdmin):
    fields = ["name"]
    inlines = [
        InlineAdminInline,
        GenericInlineAdminInline,
        IgnoredAdminInline,
        NotAnAdminInline,
    ]


@admin.register(models.Normal)
class NormalAdmin(admin.ModelAdmin):
    fields = ["name", "in_admin", "not_in_admin", "inline_admin"]


# general admin


@admin.register(models.Tag)
class TagAdmin(admin.ModelAdmin):
    fields = ["name"]


@admin.register(models.Address)
class AddressAdmin(AdminMixin, admin.ModelAdmin):
    fields = ["pk", "city", "bob", "fred", "tom", "andrew", "producer"]
    readonly_fields = ["pk", "bob", "fred", "tom", "producer"]

    def bob(self, obj):
        assert obj.street != "bad", "err"
        return "bob"

    @annotation
    def andrew(self, request, qs):
        return qs.annotate(andrew=F("street"))


class ProductMixin:
    fields = [
        "boat",
        "created_time",
        "date",
        "default_sku",
        "duration",
        "fake",
        "hidden_calculated",
        "hidden_inline",
        "hidden_model",
        "id",
        "image",
        "is_onsale",
        "model_not_in_admin",
        "name",
        "number_choice",
        "onsale",
        "producer",
        "size",
        "size_unit",
        "string_choice",
        "tags",
    ]
    readonly_fields = ["id", "is_onsale", "hidden_calculated"]
    ddb_default_filters = [
        ("a_field", "a_lookup", "a_value"),
        ("name", "not_equals", "not a thing"),
        ("a_field", "a_lookup", True),
    ]

    @attributes(ddb_hide=True)
    def hidden_calculated(self, obj):
        return obj


class ProductInline(ProductMixin, admin.TabularInline):
    model = models.Product
    ddb_hide_fields = ["hidden_inline"]
    ddb_extra_fields = ["extra_inline"]


@admin.register(models.Producer)
class ProducerAdmin(admin.ModelAdmin):
    fields = ["name", "address", "frank"]
    readonly_fields = ["frank"]
    inlines = [ProductInline]

    def frank(self, obj):
        return "frank"


class SKUInline(admin.TabularInline):
    model = models.SKU
    fields = ["name", "product", "bob"]
    readonly_fields = ["bob"]

    def bob(self, obj):  # pragma: no cover don't show funcs on inlines test
        return "bob"


@attributes(short_description="funky")
def func(obj):
    return f"f{obj.name}"


@admin.register(models.Product)
class ProductAdmin(ProductMixin, AdminMixin, admin.ModelAdmin):
    inlines = [SKUInline]
    list_display = [
        "only_in_list_view",
        "annotated",
        func,
        lambda o: f"l{o.name}",
        "calculated_boolean",
    ]
    ddb_hide_fields = ["hidden_model"]
    ddb_extra_fields = ["extra_model"]

    @annotation
    def annotated(self, request, qs):
        return qs.annotate(annotated=F("name"))

    @annotation
    def stealth_annotation(self, request, qs):
        return qs.annotate(stealth_annotation=F("name"))

    @annotation
    @attributes(ddb_hide=True)
    def hidden_annotation(self, request, qs):
        return qs.annotate(hidden_annotation=F("name"))

    @attributes(boolean=True)
    def calculated_boolean(self, obj):
        return obj.size



# Generated by Django 3.0.7 on 2020-07-02 07:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("data_browser", "0007_view_public_slug")]

    operations = [
        migrations.AddField(
            model_name="view", name="limit", field=models.IntegerField(default=1000)
        )
    ]



import itertools
import json
from collections import defaultdict

from .orm_admin import admin_get_queryset
from .orm_lookups import get_django_filter
from .query import BoundQuery
from .types import ASC, DSC


def _filter(qs, path_str, filter_):
    negation = False
    lookup = filter_.lookup

    if lookup.startswith("not_"):
        negation = True
        lookup = lookup[4:]

    filter_expression = get_django_filter(
        filter_.orm_bound_field.type_, path_str, lookup, filter_.parsed
    )
    if negation:
        return qs.exclude(filter_expression)
    else:
        return qs.filter(filter_expression)


def _cols_sub_query(bound_query):
    filters = [
        filter_
        for filter_ in bound_query.valid_filters
        if filter_.orm_bound_field.filter_
    ]

    return BoundQuery(
        bound_query.model_name, bound_query.col_fields, filters, bound_query.limit
    )


def _rows_sub_query(bound_query):
    filters = [
        filter_
        for filter_ in bound_query.valid_filters
        if filter_.orm_bound_field.filter_
    ]
    data_fields = [f for f in bound_query.data_fields if f.direction]
    return BoundQuery(
        bound_query.model_name,
        bound_query.row_fields + data_fields,
        filters,
        bound_query.limit,
    )


def get_result_queryset(request, bound_query, orm_models):
    all_fields = {f.queryset_path_str: f for f in bound_query.bound_fields}
    all_fields.update({f.queryset_path_str: f for f in bound_query.bound_filters})

    admin = orm_models[bound_query.model_name].admin
    qs = admin_get_queryset(admin, request, {f.split("__")[0] for f in all_fields})

    # sql functions and qs annotations
    for field in all_fields.values():
        qs = field.annotate(request, qs)

    # filter normal and sql function fields (aka __date)
    for filter_ in bound_query.valid_filters:
        if filter_.orm_bound_field.filter_:
            qs = _filter(qs, filter_.orm_bound_field.queryset_path_str, filter_)

    # nothing to group on, early out with an aggregate
    if not any(f.group_by for f in bound_query.bound_fields):
        return [
            qs.aggregate(
                **dict(
                    field.aggregate_clause
                    for field in bound_query.bound_fields + bound_query.bound_filters
                    if field.aggregate_clause
                )
            )
        ]

    # group by
    qs = qs.values(
        *[
            field.queryset_path_str
            for field in bound_query.bound_fields
            if field.group_by
        ]
    ).distinct()

    # aggregates
    qs = qs.annotate(
        **dict(
            field.aggregate_clause
            for field in bound_query.bound_fields + bound_query.bound_filters
            if field.aggregate_clause
        )
    )

    # having, aka filter aggregate fields
    for filter_ in bound_query.valid_filters:
        if filter_.orm_bound_field.having:
            qs = _filter(qs, filter_.orm_bound_field.queryset_path_str, filter_)

    # sort
    sort_fields = []
    for field in bound_query.sort_fields:
        if field.direction is ASC:
            sort_fields.append(field.orm_bound_field.queryset_path_str)
        if field.direction is DSC:
            sort_fields.append(f"-{field.orm_bound_field.queryset_path_str}")
    qs = qs.order_by(*sort_fields)

    return qs[: bound_query.limit]


def _get_result_list(request, bound_query, orm_models):
    return list(get_result_queryset(request, bound_query, orm_models))


def _get_objs_for_calculated_fields(request, bound_query, orm_models, res):
    # gather up all the objects to fetch for calculated fields
    to_load = defaultdict(set)
    loading_for = defaultdict(set)
    for field in bound_query.bound_fields:
        if field.model_name:
            loading_for[field.model_name].add(field.name)
            pks = to_load[field.model_name]
            for row in res:
                pks.add(row[field.queryset_path_str])

    # fetch all the calculated field objects
    objs = {}
    for model_name, pks in to_load.items():
        admin = orm_models[model_name].admin
        objs[model_name] = admin_get_queryset(
            admin, request, loading_for[model_name]
        ).in_bulk(pks)

    return objs


# dump out the results
def _format_table(fields, data, objs):
    namespace = {"objs": objs, "data": data}

    field_lines = []
    for i, field in enumerate(fields):
        namespace[f"format_{i}"] = field.get_formatter()
        value = f"row[{field.queryset_path_str!r}]"
        if field.model_name:
            value = f"objs[{field.model_name!r}].get({value})"
        field_lines.append(f"{field.path_str!r}: format_{i}({value}),")

    code = ["[None if row is None else {", *field_lines, "} for row in data]"]

    return eval("\n".join(code), namespace)


def _get_fields(row, fields):
    res = []
    for field in fields:
        v = row[field.queryset_path_str]
        if isinstance(v, list):  # pragma: postgres
            v = tuple(v)
        try:
            hash(v)
        except TypeError:
            v = json.dumps(v)
        res.append((field.queryset_path_str, v))
    return tuple(res)


def _get_data_and_all_keys(bound_query, res):
    data = defaultdict(dict)
    all_row_keys = set()
    all_col_keys = set()
    for row in res:
        row_key = _get_fields(row, bound_query.bound_row_fields)
        col_key = _get_fields(row, bound_query.bound_col_fields)
        data[row_key][col_key] = dict(_get_fields(row, bound_query.bound_data_fields))
        all_row_keys.add(row_key)
        all_col_keys.add(col_key)
    return data, all_row_keys, all_col_keys


def _get_keys(res, fields, all_keys):
    keys = {}  # abuse dict to preserve order while removing duplicates
    if fields:
        for row in res:
            key = _get_fields(row, fields)
            if key in all_keys:
                keys[key] = None
    else:
        if res:
            keys[()] = None
    return keys


def _format_grid(data, col_keys, row_keys, fields, objs):
    body_data = []
    for col_key in col_keys:
        table = []
        for row_key in row_keys:
            table.append(data[row_key].get(col_key, None))
        body_data.append(_format_table(fields, table, objs))
    return body_data


def get_results(request, bound_query, orm_models, with_format_hints):
    if not bound_query.fields:
        return {"rows": [], "cols": [], "body": []}

    res = _get_result_list(request, bound_query, orm_models)

    if bound_query.bound_col_fields and bound_query.bound_row_fields:
        # need to fetch rows and col titles seperately to get correct sort
        rows_res = _get_result_list(request, _rows_sub_query(bound_query), orm_models)
        cols_res = _get_result_list(request, _cols_sub_query(bound_query), orm_models)
    else:
        rows_res = res
        cols_res = res

    objs = _get_objs_for_calculated_fields(request, bound_query, orm_models, res)

    if bound_query.bound_col_fields or bound_query.bound_data_fields:
        data, all_row_keys, all_col_keys = _get_data_and_all_keys(bound_query, res)

        col_keys = _get_keys(cols_res, bound_query.bound_col_fields, all_col_keys)
        row_keys = _get_keys(rows_res, bound_query.bound_row_fields, all_row_keys)

        body_data = _format_grid(
            data, col_keys, row_keys, bound_query.bound_data_fields, objs
        )
        row_data = _format_table(
            bound_query.bound_row_fields, [dict(row) for row in row_keys], objs
        )
        col_data = _format_table(
            bound_query.bound_col_fields, [dict(col) for col in col_keys], objs
        )

        format_hints = {}
        for fields, data in [
            (bound_query.bound_row_fields, row_data),
            (bound_query.bound_col_fields, col_data),
            (
                bound_query.bound_data_fields,
                list(itertools.chain.from_iterable(body_data)),
            ),
        ]:
            format_hints.update(
                {field.path_str: field.get_format_hints(data) for field in fields}
            )

        return {
            "rows": row_data,
            "cols": col_data,
            "body": body_data,
            "length": len(res),
            "formatHints": format_hints,
        }
    else:
        row_data = _format_table(
            bound_query.bound_row_fields,
            [dict(_get_fields(row, bound_query.bound_row_fields)) for row in res],
            objs,
        )

        if with_format_hints:
            format_hints = {
                field.path_str: field.get_format_hints(row_data)
                for field in bound_query.bound_row_fields
            }
        else:
            format_hints = None

        return {
            "rows": row_data,
            "cols": [{}] if res else [],
            "body": [[{}] * len(res)] if res else [],
            "length": len(res),
            "formatHints": format_hints,
        }
